/*	Application		IPC sender-side using socket (UDP) - Message serialisation example	
	File			IPC_Sender_MessageSerialisation.cpp
	Purpose			Demonstration of local IPC, with a serialised data structure, using a blocking socket (on the receiving side)
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket
					2. Waits for user to type a message
					3. Sends the message to the the application port (8007) on the local computer (uses local loopback address 127.0.0.1)
					3. Exits once the message has been sent
*/

#include "stdafx.h"
#include <iostream>
#include <string>
#include <atlsocket.h>

using namespace std;

#include "IPC_Sender_MessageSerialisation.h"

void main()
{
	int iMessageLength = 0;
	m_UserInput = "";

	cout << "Local Inter-Process Communication (sender side)" << endl;
	cout << "This program waits for the user to type a message, and" << endl;
	cout << "then sends it to port 8007 on the local computer." << endl;
	cout << "------------------------------------------------------" << endl;
	cout << "Type a  message to send" << endl << endl;
	cout << "Customer name:		";
	
	// The message to be sent will be held in the form of a MyDetailsMessage structure.
	// The data values are written into the structure using the specific field names
	// (which are cName, cAddress and iAge).
	getline(cin, m_UserInput); // Get a string which may include spaces
	strncpy_s(m_SendMessage.cName,(const char*)m_UserInput.data(),20); 

	cout << "Customer address:	";
	getline(cin, m_UserInput);
	strncpy_s(m_SendMessage.cAddress,(const char*)m_UserInput.data(),50); 

	cout << "Customer age:		";
	getline(cin, m_UserInput);
	m_SendMessage.iAge = atoi(m_UserInput.data());
	cout << endl;

	// Initialise the Windows socket mechanism
	bool bSuccess;
	bSuccess = InitialiseWindowsSocketsMechanism();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the Windows socket mechanism could not be initialised" << endl;
		CloseSocketAndExit();
	}

	// Create a socket - this is the logical 'communication endpoint'
	bSuccess = CreateSocket();
	if(false == bSuccess)
	{
		cout << "Exiting the program because a socket could not be created" << endl;
		CloseSocketAndExit();
	}

	// Set the Send Address details (Use local loopback IP address, and port 8007)
	SetUpSendAddressStructFor_LocalLoopback();

	// Send the message that the user typed, to the sepcified port on the local computer
	bSuccess = SendMessage();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the message could not be sent" << endl;
		CloseSocketAndExit();
	}
	cout << "Message sent" << endl;
	
	CloseSocketAndExit();
}
	
bool InitialiseWindowsSocketsMechanism(void)
{
	
	WSADATA wsaData;
	int iResult = WSAStartup( MAKEWORD(2,2), &wsaData );  // Initialize Winsock
	if ( iResult != NO_ERROR )
	{
		cout << "Error at WSAStartup()" << endl;
		return false;
	}
	return true;
}

bool CreateSocket(void)
{
	m_SendSOCKET = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);	// Create a socket and retain a reference to it
															// AF_INET indicates that Internet addressing will be used
															// SOCK_DGRAM indicates that discrete messages will be used
															// (as opposed to a data stream)
															// PF_INET indicates that Internet protocols will be used
	if(INVALID_SOCKET == m_SendSOCKET)
	{
		cout << "socket() failed" << endl;					// Report error if socket could not be created
		return false;
	}
	return true;
}

void SetUpSendAddressStructFor_LocalLoopback(void)
{   // Initialise a SOCKADDR_IN structure with the recipients socket address details
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b1 = (unsigned  char) 127;	// Set the four bytes of internet address
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b2 = (unsigned  char) 0;		// This is the address of the recipient
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b3 = (unsigned  char) 0;		// The loopback address is used because
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b4 = (unsigned  char) 1;		// both processes will be on the same computer
	m_SendSockAddr.sin_family = AF_INET;
	m_SendSockAddr.sin_port = htons(8007);					// Set the port number. This will be used by the 
															// operating system to identify the recipient process, 
															// once the recipient has bound to the port using the
															// bind() primitive
}

bool SendMessage() // Message serialisation version
{	// The message to be sent is held in the form of a structure. The fields of the structure are
	// each of fixed length and the structure is held in contiguous memory (there are no gaps
	// between fields). Therefore, serialisation only requires that the address of the beginning 
	// of the structure '&m_SendMessage', and the size of the structure 'sizeof(m_SendMessage)'
	// are passed as parameters to the sendto() primitive, so that the structure itself is used
	// as the message buffer. The contents of the structure are sent as a stream of bytes without
	// details of the fields they belong to.

	// Send the user's typed message (which is in the string variable m_UserInput)
	// The sendto() primitive uses the m_SendSOCKET reference to identify the socket in use,
	// and the m_SendSockAddr address structure to identify where to send the message
	int iBytesSent = sendto(m_SendSOCKET, (char FAR *)&m_SendMessage, sizeof(m_SendMessage), 0,
			(const struct sockaddr FAR *)&m_SendSockAddr, sizeof(m_SendSockAddr));
	if(INVALID_SOCKET == iBytesSent)
	{
		cout << "sendto() Failed!" << endl;		// Report error if send was unsuccessful
		return false;
	}
	return true;
}

void CloseSocketAndExit()
{
	char CUserResponse;
 	if(0 != m_SendSOCKET)
	{
		closesocket(m_SendSOCKET);	// If the socket was open, Close the socket
	}
	WSACleanup();	
	cout << "Socket closed" << endl << endl;
	cout << "(press a key to continue)" << endl;
	cin >> CUserResponse;	// Wait for user to type a key (to allow them to read the message before the window is closed)
	exit(0);
}